/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file audio.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 25.02.2006
 */

#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>

#include "common.h"		/** common project settings */
#include "audio.h"		/** common for all audio */
#include "mp3.h"		/** mp3 decoder */
#include "aac.h"		/** aac decoder */
#include "parsetime.h"	/** converts time-formats */
#include "report.h"		/** reporter class */


/*
 * Global vars
 * TODO: use malloc instead?
 */
static struct audio_data audio_data;			/** makes a struct */
static struct audio_data *data = &audio_data;	/** and a pointer to reference it */


/**
 * Local function declarations
 */
static
bool open_file (const char *fname, struct audio_data *data);


static
audio_file_type get_file_type (const char *file);


int play (const char *fname, char *start, char *stop)
{
	if (DEBUG_AUDIO) report ("playing file. . ", REP_DEBUG);
	/** initialize some data in the audio_data struct */
	data->progress.seconds = 0;		data->progress.fraction = 0;
	data->start	=	parse_time (start);
	data->stop	=	parse_time (stop);

	/** opening and mmapping a new file (if we don't
	 * already have the file loaded and mmapped) 
	 * If an error occurs, open_file will return false and
	 * the file in question will be skipped. */
	if (open_file (fname, data) == false)
	{
		/** constructing and reporting an error-message */
		char message[STRLEN];
		snprintf (message, sizeof (message), "Could not open file. (%s)", fname);
		report (message, REP_WARNING);
		return false;
	}

	/** find out what sort of file were dealing with */
	audio_file_type ft = get_file_type(fname);
	switch (ft)
	{	
		case (AUDIO_FILE_AAC):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is aac", REP_DEBUG);
			if (data->prev_file_type != AUDIO_FILE_AAC)	/** last file_type was NOT aac */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT aac", REP_DEBUG);
				/** close decoder and start a new one */
				mp3_close (data);
				//aac_init (data);
			}
			/** decode data */
			//aac_play(data);
			break;
		case (AUDIO_FILE_MP3):
			if (DEBUG_AUDIO) report ("audio.c: audio-file is mp3", REP_DEBUG);
			if (data->prev_file_type != AUDIO_FILE_MP3) /** last file_type was NOT mp3 */
			{
				if (DEBUG_AUDIO) report ("the last filetype was NOT mp3", REP_DEBUG);
				/** close decoder and start a new one */
				//aac_close (data);
				mp3_init (data);
			}
			/** decode data */
			mp3_play(data);
			break;
		case (AUDIO_FILE_WAV):
			report ("wav parsing isn't implemented yet", REP_MESSAGE);
			break;
		case (AUDIO_FILE_UNKNOWN):
		default:		/** This should never happen! */
			assert(0);	/** Report where we are and exit the program */
			break;
	}
	return false;
}


/**
 * This function will load a file if it's not already loaded.
 * The freeing and generating of mmaps are handled here.
 * Returns -1 on error, 1 on success. */
static
bool open_file (const char *fname, struct audio_data *data)
{
	if (DEBUG_AUDIO) {
		char rep[STRLEN];
		snprintf (rep, sizeof (rep), "audio.c asked to open file: %s\t(data->filename is: %s)", fname, data->filename);
		report (rep, REP_DEBUG);
	}

	/** did we already load the file? */
	/** We have to check if data->filename is NULL /first!/ If one of
	 * strcmp's arguments are NULL, we've made a booboo */
	if ((data->filename != NULL) && !strcmp (data->filename, fname))
	{
		if (DEBUG_AUDIO) report ("File already loaded", REP_MESSAGE);
	}

	/** if not, we'll need to free the old mmap, open the new
	 * file and create a new mmap */
	else
	{
		/* Free mmap, clean data */
		if (data->fdm != NULL) {

			/** freeing the mmap before creating a new one */
			if (munmap (data->fdm, data->stat.st_size) == -1)
				report ("Unable to free mmap", REP_CRITICAL);

			if (DEBUG_AUDIO) {
				char rep[STRLEN];
				snprintf (rep, sizeof (rep), "Freed mmap. Pointer: %p, Size: %d", data->fdm, (int) data->stat.st_size);
				report (rep, REP_DEBUG);
			}

			data->fdm = NULL;		/** we have freed the data - don't want an old pointer to point there! */
			data->filename = NULL;	/** this file is no longer in our memory */
			close (data->fd);		/** closing the file before we open a new one */
		}
		
		/** udating the data-struct with the new filename */
		if (data->filename != NULL)
			free(data->filename);
		data->filename = (char *) malloc (strlen(fname)+1);
		if (DEBUG_AUDIO) report ("allocating room for data->filename", REP_DEBUG);
		strcpy(data->filename, fname);
		if (DEBUG_AUDIO) report ("copying fname to data->filename", REP_DEBUG);
		char msg[STRLEN];
		snprintf (msg, sizeof (msg), "data->filename is: %s\tfname is: %s", data->filename, fname);
		if (DEBUG_AUDIO) report (msg, REP_DEBUG);
		
		
		/** opening new file */
		if ((data->fd = open (fname, O_RDONLY)) == -1)
		{
			report ("Error opening file", REP_WARNING);
			return false;
		}

		/** setting filetype */
		data->prev_file_type = data->file_type;
		data->file_type = get_file_type(fname);
		if (DEBUG_AUDIO) 
		{
			char msg[STRLEN];
			snprintf (msg, sizeof (msg), "Filename %s is of type: %d", data->filename, data->file_type);
			report(msg, REP_DEBUG);
		}

		/** fstating file. this gives us alot of information about
		 * the file (we'll use the st_size (filesize) here) */
		if (stat(fname, &data->stat) == -1 || data->stat.st_size == 0)
		{
			report ("Error fstating file or filesize is 0", REP_WARNING);
			return false;
		}

		/** mmaps the file */
		data->fdm = mmap (0, data->stat.st_size, PROT_READ, MAP_SHARED, data->fd, 0);
		if (data->fdm == MAP_FAILED)
		{
			report ("Error mmapping file", REP_WARNING);
			return false;
		}
		
		if (DEBUG_AUDIO)
		{
			char rep[STRLEN];		
			snprintf (rep, sizeof (rep), "mmapped new file. Pointer: %p, size: %d", data->fdm, (int) data->stat.st_size);		
			report (rep, REP_DEBUG);
		}
	}
	return true;
}



/*
 * Finds out what filetype a file is (mp3, aac, wav).
 */
static
audio_file_type get_file_type (const char *fname)
{
//	if (DEBUG_AUDIO) report("Querying file_type. Returning mp3 for now.", REP_DEBUG);
	return AUDIO_FILE_MP3;
}
