/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file control.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 * Datastructure and methods for controlling it.
 */

 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "control.h"
#include "nccsmilparser.h"
#include "ncxsmilparser.h"
#include "report.h"

#ifdef LIBXML_READER_ENABLED

// static function declarations
static void initNode (struct Node* node);
static void initSmilNode (struct SmilNode* smilNode);
static void flushSmilNodes ();
static void flushNodes ();


/**
 * Initializes the data structure for the parsing (linked lists)
 */
void parseInit ()
{
	 /*
     * this initialize the library and check potential ABI mismatches
     * between the version it was compiled for and the actual shared
     * library used.
     */
	LIBXML_TEST_VERSION
	
	/* initializing the head and the tail of the linked lists */
	smilHead = (struct SmilNode *) malloc (sizeof(struct SmilNode));
	initSmilNode (smilHead);
	smilTail = (struct SmilNode *) malloc (sizeof(struct SmilNode));
	initSmilNode (smilTail);
	smilHead->next = smilTail;
	smilTail->prev = smilHead;
	smilPos = smilHead;
	
	nodeHead = (struct Node *) malloc (sizeof(struct Node));
	initNode (nodeHead);
	nodeTail = (struct Node *) malloc (sizeof(struct Node));
	initNode (nodeTail);
	nodeHead->next = nodeTail;
	nodeTail->prev = nodeHead;
	nodePos = nodeHead;
}


/**
 * Method to free all memory allocated by the parser. Terminates all parsing
 * functionality.
 */
void parseTerminate ()
{
	flushNodes();
	flushSmilNodes();
	
	/*
	* Cleanup function for the XML library.
	*/
	xmlCleanupParser ();
    
	/*
	 * this is to debug memory for regression tests
	*/
	xmlMemoryDump ();
	
	free (nodeHead);
	free (nodeTail);	
}


/**
 * Removes whitespaces, newlines and tabs before and after a sentence.
 *
 * @param string contains the string to check.
 *
 * Return the new char* if the sentence contains any characters,
 * NULL otherwise
 */
char* removewhitespaces(char* string)
{	
	int start = 0; // Counter variables	
	int end = 0;
	int i = 0;
	int n = 0;
	char* newstring = NULL; // The new string
	
	/* Checks the characters before the string */
	for (start = 0; start < strlen(string); start++)
	{	
		if (string[start] != ' ' && string[start] != '\t' && string[start] != '\n')
			break;
	}
	
	/* If the sentences only contains whitespaces,tabs or newlines, return NULL */
	if (start == strlen(string)) return NULL;
	
	/* Checks the characters from the end and backwards */
	for (end = strlen(string)-1; end >= 0; end--)
	{	
		if (string[end] != ' ' && string[end] != '\t' && string[end] != '\n')
			break;
	}
	
	/* Creates the new string with the non-characters stripped */	
	newstring = malloc(end-start+2);
	for (i = start; i <= end; i++)
	{
		newstring[n++] = string[i];
	}
	newstring[n] = '\0';	
	return newstring;
}


/**
 * Converts min:sec.frac to sec.frac
 * 
 * @return the string contianing the offset
 */
char* getTime(char* time)
{
	char* tmp, *min, *sec, *frac = NULL;
	char* value = (char *) malloc(6);	
	int minutes = 0, seconds = 0;
	
	// copy time to a tmp variable
	tmp = (char *) malloc (strlen (time)+1);
	strcpy (tmp, time);
	// tokenize in order to get minutes
	min = strtok (tmp, ":");
	minutes = atoi (min);
	// tokenize in order to get the chars after min:
	tmp = strtok (NULL, ":");
	// tokenize in order to get seconds
	sec = strtok (tmp, ".");
	// convert char to int in order to add minutes to seconds
	seconds = atoi(sec);
	seconds = (minutes * 60) + seconds;
	// tokenize in order to get fraction
	frac = strtok (NULL, ".");	
	// convert the int seconds to char
	snprintf(value, sizeof (value), "%d", seconds);
	// add .fractions
	strcat (value, ".");
	strcat (value, frac);
//	free (tmp);
	return value;
}


/**
 * Adds a new SmilNode at the end of the linked list
 */
struct SmilNode* addNewSmilNode ()
{
	// allocating memory for the new smilnode
	struct SmilNode *smilNode = (struct SmilNode *) malloc (sizeof(struct SmilNode));
	
	// initializes the smilNode
	initSmilNode (smilNode);
	
	// adding the empty smilnode til the end of the linked list
	struct SmilNode *tmp = smilTail->prev;
	tmp->next = smilNode;
	smilNode->prev = tmp;
	smilNode->next = smilTail;
	smilTail->prev = smilNode;
	
	return smilNode;
}

/**
 * Sets the smilNodes pointers to NULL and all values to 0
 */
static void initSmilNode (struct SmilNode* smilNode)
{
	smilNode->anchor = NULL;
	smilNode->fragmentIdentifier = NULL;
	smilNode->header = NULL;
	smilNode->next = NULL;
	smilNode->prev = NULL;
	smilNode->optional = 0;
}

 
 /**
  * Adds a new Node at the end of the linked list
  */
struct Node* addNewNode ()
{
	// allocating memory for the new smilnode
	struct Node *node = (struct Node *) malloc (sizeof(struct Node));
	
	// initializes the node
	initNode(node);
	
	// adding the empty smilnode til the end of the linked list
	struct Node *tmp = nodeTail->prev;
	tmp->next = node;
	node->prev = tmp;
	node->next = nodeTail;
	nodeTail->prev = node;
	
	return node;
}

/**
 * Sets the nodes pointers to NULL and all values to 0
 */
static void initNode (struct Node* node)
{
	node->audiofilename = NULL;
	node->audioStartPos = NULL;
	node->audioStopPos = NULL;
	node->next = NULL;
	node->prev = NULL;
	node->textPassage = NULL;
}


/**
 * Removes all smilnodes in the linked list and deallocates their memory.
 */
static void flushSmilNodes ()
{
	struct SmilNode *tmp = smilHead->next;
	while (tmp != smilTail)
	{
		if (tmp->anchor != NULL)
			free (tmp->anchor);
		if (tmp->fragmentIdentifier != NULL)
			free (tmp->fragmentIdentifier);
		if (tmp->header != NULL)
			free (tmp->header);
		tmp = tmp->next;
		free (tmp->prev);
	}
	smilHead->next = smilTail;
	smilTail->prev = smilHead;
	smilPos = smilHead;
}


/**
 * Removes all nodes in the linked list and deallocates their memory.
 */
static void flushNodes ()
{
	struct Node *tmp = nodeHead->next;
	while (tmp != nodeTail)
	{
		if (tmp->audiofilename != NULL)
			free (tmp->audiofilename);
		if (tmp->audioStartPos != NULL)
			free (tmp->audioStartPos);
		if (tmp->audioStopPos != NULL)
			free (tmp->audioStopPos);
		if (tmp->textPassage != NULL)
			free (tmp->textPassage);
		tmp = tmp->next;
		free (tmp->prev);	
	}
	nodeHead->next = nodeTail;
	nodeTail->prev = nodeHead;
	nodePos = nodeHead;
}


/**
 * Function for seeking in the data structure. It can seek paragraphs (nodes) and
 * chapters (smilnode).
 * 
 * @param type -2 = prev para, -1 = prev chap, 1 = next chap, 2 = next para
 * @param book 0 = ncc (2.02), 1 = ncx (Z39.86-2005)
 * @returns 0 if end of book, 1 if success and -1 on error
 */
int seek (int type, int book)
{
	int ret = 1;
	
	if (smilHead->next == smilTail) return 0;
	
	switch (type)
	{
		case -2:
			if (nodePos == nodeHead || nodePos == nodeHead->next)
			{
				if (smilPos == smilHead)
				{
					smilPos = smilHead->next;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(smilPos);	
					else
						ret = parseNCXSmil(smilPos);
					if (ret != 1) return ret;
					nodePos = nodeHead->next;
					return ret;	
				}
				else if (smilPos == smilHead->next)	
				{
					nodePos = nodeHead->next;
					return ret;	
				}
				else
				{
					smilPos = smilPos->prev;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(smilPos);	
					else
						ret = parseNCXSmil(smilPos);
					if (ret != 1) return ret;
					nodePos = nodeTail->prev;
					return ret;	
				}
			}
			else
			{
				nodePos = nodePos->prev;
				return ret;
			}
		
		case -1:
			if (smilPos == smilHead)
			{
				smilPos = smilHead->next;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(smilPos);	
				else
					ret = parseNCXSmil(smilPos);
				if (ret != 1) return ret;
				nodePos = nodeHead->next;
				return ret;
			}
			else if (smilPos == smilHead->next)
			{
				nodePos = nodeHead->next;
				return ret;	
			}
			else
			{
				smilPos = smilPos->prev;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(smilPos);	
				else
					ret = parseNCXSmil(smilPos);
				if (ret != 1) return ret;
				nodePos = nodeHead->next;
				return ret;	
			}
		
		case 1:
			if (smilPos == smilTail)
			{
				return 0;
			}
			else if (smilPos == smilTail->prev)
			{
				smilPos = smilTail;
				return 0;
			}
			else {
				smilPos = smilPos->next;
				flushNodes();
				if(book == 0)
					ret = parseNCCSmil(smilPos);	
				else
					ret = parseNCXSmil(smilPos);
				if (ret != 1) return ret;
				nodePos = nodeHead->next;
				return ret;
			}
		
		case 2:
			if (nodePos == nodeTail || nodePos == nodeTail->prev)
			{
				if (smilPos == smilTail)
				{
					nodePos = nodeTail;
					return 0;
				}
				else if (smilPos == smilTail->prev)
				{
					smilPos = smilTail;
					nodePos = nodeTail;
					return 0;	
				}
				else
				{
					smilPos = smilPos->next;
					flushNodes();
					if(book == 0)
						ret = parseNCCSmil(smilPos);	
					else
						ret = parseNCXSmil(smilPos);
					if (ret != 1) return ret;
					nodePos = nodeHead->next;
					return ret;
				}
			}
			else
			{
				nodePos = nodePos->next;
				return ret;	
			}
		
		default:
			return -1;	
		
	}
}


#endif
