/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file mp3.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 15.02.2006
 * 
 * This code uses libmad0 to do mp3 decoding, and libao to do the audio output.
 * It borrows code from / are inspired by mpg321 and minimad.c.
 */
 
// * * Resources:
// * * * A good description of how mad works (the closest thing to an API we'll get?).
// * * * * http://trac.videolan.org/vlc/file/tags/v0_3_0/plugins/mad/API?rev=6952
// * * * http://www.mars.org/mailman/public/mad-dev/2001-August/000338.html
// * * * http://www.mars.org/mailman/public/mad-dev/2001-May/000244.html


 /*
  * includes
  */
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <mad.h>

#include "common.h"		/** common project settings */
#include "mp3.h"
#include "daisylibao.h"
#include "audio.h"
#include "report.h"
#include "parsetime.h"


/*
 * Local function declarations
 */
static int mp3_decode (struct audio_data *);


/**
 * this will need t be called before we're using mp3-decoder
 * in order to initialize it.
 */
void mp3_init (struct audio_data *data)
{
	/** Initializing the mp3 data struct */
	if (data == NULL) { report ("data pointer is NULL.", REP_CRITICAL); }
	if (DEBUG_MP3) report("initializing audio_data", REP_DEBUG);
	data->progress.fraction = 0;
	data->progress.seconds  = 0;
	//data->filename = NULL;
	//data->fdm = NULL;
}


/**
 * this will need to be called once we are done using the mp3-decoder
 * in order to free resources.
 */
void mp3_close (struct audio_data *data)
{
	
	/** Freeing daisyLibao */
	libao_close ();
	
	/* Freeing mmap */
	if (munmap (data->fdm, data->stat.st_size) == -1)
		report ("Error freeing mmap", REP_MESSAGE);

	/** closing the file */
	close (data->fd);
}


/**
 * dither struct.
 */
struct audio_dither {
  mad_fixed_t error[3];
  mad_fixed_t random;
};


/**
 * 32-bit pseudo-random number generator
 * This is used to do dithering.
 */
static inline
unsigned long prng (unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}


/**
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. Right now, an entire file has
 * been memory-mapped. We just call mad_stream_buffer() with the
 * address and length of the mapping.
 */
static
enum mad_flow input (void *audio_data,
					 struct mad_stream *stream)
{
	/** contains information about the file we're working on,
	 * start-, stop- and progress-times etc */
	struct audio_data *data = audio_data;
	
	if (!data->buffer.buffer_length)
		return MAD_FLOW_STOP;

	mad_stream_buffer (stream, data->buffer.buffer_start, data->buffer.buffer_length);

	data->buffer.buffer_length = 0;

	return MAD_FLOW_CONTINUE;
}


/**
 * generic linear sample quantize and dither routine
 */
static
inline
signed long audio_linear_dither (unsigned int bits, mad_fixed_t sample,
                                 struct audio_dither *dither)
{
	unsigned int scalebits;
	mad_fixed_t output, mask, random;

	enum {
		MIN = -MAD_F_ONE,
		MAX =  MAD_F_ONE - 1
	};

	/* noise shape */
	sample += dither->error[0] - dither->error[1] + dither->error[2];

	dither->error[2] = dither->error[1];
	dither->error[1] = dither->error[0] / 2;

	/* bias */
	output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

	scalebits = MAD_F_FRACBITS + 1 - bits;
	mask = (1L << scalebits) - 1;

	/* dither */
	random  = prng (dither->random);
	output += (random & mask) - (dither->random & mask);

	dither->random = random;

	/* clip */
	if (output > MAX) {
		output = MAX;

		if (sample > MAX)
			sample = MAX;
	}
	else if (output < MIN) {
		output = MIN;

		if (sample < MIN)
			sample = MIN;
	}

	/* quantize */
	output &= ~mask;

	/* error feedback */
	dither->error[0] = sample - output;

	/* scale */
	return output >> scalebits;
}


/**
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to prepare the decoded PCM audio and tell libao to play it.
 * (Some of the folowing code are borowed from mpg321)
 */
static
enum mad_flow output (void *audio_data,
				      struct mad_header const *header,
				      struct mad_pcm *pcm)
{
	//TODO: This was never used //struct audio_data *data = audio_data;

	/** Buffer that we will tell libao to play. 1152 is the max buffer
	 * we will get from mad. Stereo and a 16bit samplesize
	 * means that we'll  have to multiply it by 4 (2 channels * 2 bytes).
	 */
	static unsigned char stream[1152 * 4];
	
	/** A pointer within the stream-buffer.
	 * uses register keyword because we'll probably use it alot */
	register unsigned char *ptr = stream;

	/** This contains our current sample. this is what we let libao play */
	register signed int sample;
	
	/** This contains our temporary sample.
	 * we will run audio_linear_dither and normalize this */
    register mad_fixed_t tempsample;
    
    static struct audio_dither dither;
	
	unsigned int nchannels, nsamples;
	mad_fixed_t const *left_ch, *right_ch;

	nchannels = pcm->channels;
	nsamples  = pcm->length;
	left_ch   = pcm->samples[0];
	right_ch  = pcm->samples[1];
	static int samplerate;
	samplerate = pcm->samplerate;


	/* We need to know information about the file before we can open the playdevice
	 * in some cases. So, we do it here. */
	if (!ao_device_is_null())	// if ao haven't opened a playdevice yet
	{
//		nchannels = MAD_NCHANNELS(header);
//		samplerate = header->samplerate;
		if (DEBUG_MP3) report ("First time we play - initializing daisylibao", REP_DEBUG);
		libao_init(AO_FMT_LITTLE, pcm->channels, pcm->samplerate);
//		if (DEBUG_MP3)
//		{
//			char msg[STRLEN];
//			report ("debug info.", REP_DEBUG);
//			snprintf (msg, STRLEN, "\tbits: %lu\n\tchannels: %d\n\trate: %d", header->bitrate, pcm->channels, pcm->samplerate);
//			report (msg, REP_DEBUG);
//		}
	}
	/** if channels or samplerate changes. (this shouldn't happen in daisy-books) */
	else if ((nchannels != MAD_NCHANNELS(header) || samplerate != header->samplerate))
	{
		if (DEBUG_MP3) report ("channels or samplerate have changed!", REP_DEBUG);
		//TODO: report this as debuging? //printf ("\nnch: %d, MNC: %d", nchannels, MAD_NCHANNELS(header));
		//TODO: report this as debuging?//printf ("\nsamplra: %d, head->srt: %d", samplerate, header->samplerate);
		libao_close();
		libao_init(AO_FMT_LITTLE, pcm->channels, pcm->samplerate);
	}


	/** i imagine header->samplerate and pcm->samplerate always are the same. but, I'm curious, so let's check. */
	if (DEBUG_MP3)
	{
		if (header->samplerate != pcm->samplerate)
			report ("Wow! - header->samplerate are difrent from pcm->samplerate", REP_DEBUG);
	}


	/* Stereo
	 * output sample(s) in 16-bit signed little-endian PCM */
	if (pcm->channels == 2)
	{
		while (nsamples--) {
			tempsample = *left_ch++;
			sample = (signed int) audio_linear_dither(16, tempsample, &dither);
			
			/* Assumes that words ar big-endian. If they're not, we'll need to switch the two following lines. */
			*ptr++ = (unsigned char) (sample >> 0);
			*ptr++ = (unsigned char) (sample >> 8);
			
			
			tempsample = *right_ch++;
			sample = (signed int) audio_linear_dither(16, tempsample, &dither);
			
			/* Assumes that words ar big-endian. If they're not, we'll need to switch the two following lines. */
			*ptr++ = (unsigned char) (sample >> 0);
			*ptr++ = (unsigned char) (sample >> 8);
		}
		libao_play (stream, pcm->length * 4);
	}

	/* mono
	 * output sample(s) in 16-bit signed little-endian PCM */
	else
	{
		while (nsamples--)
		{
			tempsample = *left_ch++;
			sample = (signed int) audio_linear_dither(16, tempsample, &dither);
			
			/* Assumes that words ar big-endian. If they're not, we'll need to switch the two following lines. */
			*ptr++ = (unsigned char) (sample >> 0);
			*ptr++ = (unsigned char) (sample >> 8);
		}
		libao_play (stream, pcm->length * 2);
	}
	/** Continue decoding */
	return MAD_FLOW_CONTINUE;
}

/**
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static
enum mad_flow error(void *audio_data,
				    struct mad_stream *stream,
				    struct mad_frame *frame)
{
	struct audio_data *data = audio_data;

	/** prepare an error-message */
	char error[STRLEN];
	snprintf (error, sizeof (error), "decoding error 0x%04x (%s) at byte offset %u",
			 stream->error, mad_stream_errorstr(stream),
			 stream->this_frame - data->buffer.buffer_start);

	/** report the error-message */
//	report (error, REP_WARNING);

	/* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

	return MAD_FLOW_CONTINUE;
}


/**
 * This is the libao_play-method that should be called by the xml-parser.
 * This is commented out and moved to audio.c.
 */
bool mp3_play (struct audio_data *data)
{
	/** initialize some data in the audio_data struct */
	data->progress.seconds = 0;
	data->progress.fraction = 0;

	if (DEBUG_MP3)
	{
		char msg[STRLEN];
		snprintf (msg, sizeof (msg), "mp3_play: Starting mp3 decoding.\ndata->filename is: %s\t data->fdm is: %p", data->filename, data->fdm);
		report(msg, REP_DEBUG);
	}

	/** start decoding the data */
	mp3_decode (data);
	return false;
}


/**
 * header analyses the frame header and do things like seeking.
 * fraction is of type unsigned long
 * seconds is of type long
 */ 
static
enum mad_flow header (void *audio_data, struct mad_header const *header)
{
	struct audio_data *data = audio_data;
	
	/** we increment our duration */
	data->progress.fraction += header->duration.fraction;
	data->progress.seconds  += header->duration.seconds;

	if (data->progress.fraction > MAD_TIMER_RESOLUTION)
	{
		data->progress.seconds  += data->progress.fraction / MAD_TIMER_RESOLUTION;
		data->progress.fraction %= MAD_TIMER_RESOLUTION;
	}
	
	/** compares progress agains start- and stop-points */
	if (data->start.seconds > data->progress.seconds)
		return MAD_FLOW_IGNORE;
		
	if (data->start.seconds == data->progress.seconds && data->start.fraction > data->progress.fraction)
		return MAD_FLOW_IGNORE;
	
	if ((data->stop.seconds <= data->progress.seconds) &&
		(data->stop.fraction <= data->progress.fraction))
	{
		return MAD_FLOW_STOP;
	}

	return MAD_FLOW_CONTINUE;
}


/**
 * This is the function called by main() above to perform all the decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, header, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 */
static
int mp3_decode (struct audio_data *data)
{
	struct mad_decoder decoder;
	int result;
	
	/** initialize the buffer-part of the audio_data struct */
	data->buffer.buffer_start = data->fdm;
	data->buffer.buffer_length = data->stat.st_size;
	
	/** configure input, output, and error functions */
	mad_decoder_init (&decoder, data,
		   input, header, 0 /* filter */, output,
		   error, 0 /* message */);

	/** start decoding */
	result = mad_decoder_run (&decoder, MAD_DECODER_MODE_SYNC);

	/** release the decoder */
	mad_decoder_finish (&decoder);

	return result;
}
