/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file nccsmilparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 * A parser for the SMIL 1.0 standared used with Daisy 2.02 books. 
 * Collects data and stores it in linked list.
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "nccsmilparser.h"
#include "control.h"
#include "report.h"
#include "common.h"

#ifdef LIBXML_READER_ENABLED

// static function declarations 
static int parseSmilHead (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilLayout (xmlTextReaderPtr reader);
static int parseSmilMeta (xmlTextReaderPtr reader);
static int parseSmilRegion (xmlTextReaderPtr reader);
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilMainSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilNestedSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode, struct Node *tmpNode);
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseXHTML (char* filename, char* fragment, struct Node *tmpNode);
static int parseXHTMLHead (xmlTextReaderPtr reader);
static int parseXHTMLTitle (xmlTextReaderPtr reader);
static int parseXHTMLMeta (xmlTextReaderPtr reader);
static char* parseXHTMLHtmltag (xmlTextReaderPtr reader);
static int parseXHTMLBody (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode);


// global variables
static int alreadyfound = 0; //Used in the smil-parsing
extern char *path;


/************************ Code for parsing a *.smil files ************************/


/**
 * Method for parsing of a *.smil file
 * 
 * @param tmpSmilNode - the smilNode for the smilfile to parse.
 * @returns 1 in case of success and -1 in case of error
 */
int parseNCCSmil (struct SmilNode *tmpSmilNode)
{
 	const xmlChar *name = NULL;
 	xmlTextReaderPtr reader = NULL;
 	int ret = 1;
 	
 	char tmp[STRLEN];
	strcpy(tmp, path);
	strcat(tmp, "/");
	strcat(tmp, tmpSmilNode->anchor);
	
 	// open the smil file
 	reader = xmlReaderForFile (tmp, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open SMIL file: %s", tmpSmilNode->anchor);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCCSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL file: %s", tmpSmilNode->anchor);
	 		report (error, REP_DEBUG);
		}
		
	 	// read the first tag
	 	ret = xmlTextReaderRead(reader);
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			xmlFreeTextReader (reader);
 			return -1;
	 	}
	 	// while there are more tags, parse the next
		while(ret == 1)
	 	{
	 		// get name of tag
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
			//parse tag if <head>
			if (!strcmp (name, "head"))
			{ 
				ret = parseSmilHead (reader, tmpSmilNode);
				if (ret != 1)
	 			{
	 				char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			xmlFreeTextReader (reader);
		 			return -1;
			 	}
			}
			//parse tag if <body>
			else if (!strcmp (name, "body"))
			{ 
				ret = parseSmilBody (reader, tmpSmilNode);
				if (ret != 0)
			 	{
			 		char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			xmlFreeTextReader (reader);
		 			return -1;
			 	}
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return 1;
}


/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilHead(xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <HEAD>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	// while we haven't come to the head closetag
	while (strcmp (name, "head"))
	{	
		//parse tag if <layout>
		if (!strcmp (name, "layout"))
		{
			ret = parseSmilLayout (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <meta>
		else if (!strcmp (name, "meta"))
		{
			ret = parseSmilMeta (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Layout-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilLayout (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <LAYOUT>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	//parse tag if <#text>
	if (!strcmp(name, "#text")) 
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	//parse tag if <region>
	if (!strcmp(name, "region")) 
	{
		ret = parseSmilRegion (reader);
		if (ret != 1) return ret;
	}
	ret = xmlTextReaderRead (reader);
	return ret;
}

/**
 * Parse Meta-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilMeta (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Region-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilRegion (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <REGION>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <BODY>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	//parse tag if <#text>
	if (!strcmp(name, "#text")) 
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	//parse tag if <seq>
	if (!strcmp (name, "seq")){
		ret = parseSmilMainSeq (reader, tmpSmilNode);
		if (ret != 1) return ret;
	}
	ret = xmlTextReaderRead (reader);	
	return ret;
}

/**
 * Parse Main-Seq-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilMainSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL main <SEQ>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while the current node is par seq or #text
	while (!strcmp (name, "par") ||  !strcmp (name, "seq") || !strcmp (name, "#text"))
	{	
		//parse tag if <par>
		if (!strcmp (name, "par")){
			ret = parseSmilPar (reader, tmpSmilNode);
			if (ret != 1) return ret;
		}
		//parse tag if <seq>
		else if (!strcmp (name, "seq"))
		{
			ret = parseSmilNestedSeq (reader, tmpSmilNode, NULL);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Par-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	xmlChar *tmp = NULL;
	int ret;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <PAR>", REP_DEBUG);
	// go to the fragment identifier point and parses from there.
	tmp = xmlTextReaderGetAttribute (reader, "id");
	// if the par tag has an id attribute, read it and check wheter it matches the fragmentid.
	// if there is none id attribute, set alreadyfound to 1 and parse the whole smilfile
	if (tmp != NULL) 
	{
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
		tmp = xmlTextReaderGetAttribute (reader, "id");
		if (!strcmp (tmp, tmpSmilNode->fragmentIdentifier))
			alreadyfound = 1;
	}
	else
		alreadyfound = 1;
	if (tmp != NULL) xmlFree (tmp); tmp = NULL;
	if (alreadyfound == 1)
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		
		struct Node *tmpNode = addNewNode();
		
		// while we haven't come to the par closetag
		while (strcmp (name, "par"))
		{	
			//parse tag if <text>
			if (!strcmp (name, "text"))
			{
				ret = parseSmilText (reader, tmpNode);
				if (ret != 1) return ret;
			}	
			//parse tag if <audio>
			else if (!strcmp (name, "audio"))
			{
				ret = parseSmilAudio (reader, tmpNode);
				if (ret != 1) return ret;
			}
			//parse tag if <seq>
			else if (!strcmp (name, "seq"))
			{
				ret = parseSmilNestedSeq (reader, tmpSmilNode, tmpNode);
				if (ret != 1) return ret;
			}
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
		}
	}
	else 
	{
		ret = xmlTextReaderNext (reader);	
	}
	return ret;
}

/**
 * Parse Nested-Seq-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilNestedSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode, struct Node *tmpNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <SEQ>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't come to the seq closetag
	while (strcmp (name, "seq"))
	{	
		//parse tag if <par>
		if (!strcmp (name, "par"))
		{
			ret = parseSmilPar (reader, tmpSmilNode);
			if (ret != 1) return ret;
		}
		//parse tag if <audio>
		else if (!strcmp (name, "audio"))
		{
			if (tmpNode->audiofilename != NULL)
				tmpNode = addNewNode();
			ret = parseSmilAudio (reader, tmpNode);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);	
		if (name == NULL) return 0;
	}
	return ret;
}

/**
 * Parse Text-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *textsrc, *fragment, *attr;
	textsrc = fragment = attr = NULL;
	const xmlChar *name = NULL;
	char *tmp = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <TEXT>", REP_DEBUG);
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	if (!strcmp(name, "#text")) {
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
	}
	attr = xmlTextReaderGetAttribute(reader, "src");	
	
	// splits the text on #.
	// get the anchor url
	textsrc = strtok (attr, "#");
	
	if (DEBUG_NCCSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> XMLFile: %s", textsrc);
		report (error, REP_DEBUG);
	}
	
	// storing the filename of the text file
	tmp = (char *) malloc (strlen (textsrc)+1);
	strcpy (tmp, textsrc);
	tmpNode->textfilename = tmp;
	
	// get the fragment identifier
	fragment = strtok (NULL, "#");
	
	if (DEBUG_NCCSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> Fragment Identifier: %s", fragment);
		report (error, REP_DEBUG);
	}
	
	// storing the fragment identifier for the passage in the text file
	tmp = (char *) malloc (strlen (fragment)+1);
	strcpy (tmp, fragment);
	tmpNode->fragmentidentifier = tmp;
	
	// parse text file
	if (parseXHTML(textsrc, fragment, tmpNode) != 0)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Failed to parse XHTML file: %s", tmpNode->textfilename);
 		report (error, REP_WARNING);
 		return -1;
	}
	if (attr != NULL) xmlFree (attr); attr = NULL;
	return ret;
}

/**
 * Parse Audio-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *attr, *value;
	attr = value = NULL;
	char *tempChar = NULL;
	int ret = 1;
	
	if (DEBUG_NCCSMILPARSER) report ("Parsing SMIL <AUDIO>", REP_DEBUG);
	// fetch the url anchor to the audio file
	attr = xmlTextReaderGetAttribute (reader, "src");
	tempChar = (char *) malloc (strlen(attr)+1);
	strcpy(tempChar, attr);
	
	if (DEBUG_NCCSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> Audiofile: %s", attr);
		report (error, REP_DEBUG);
	}
	
	// storing the audio filename
	tmpNode->audiofilename = tempChar;	
	if (attr != NULL) xmlFree (attr); attr = NULL;
	
	// fetch clip-begin
	attr = xmlTextReaderGetAttribute (reader, "clip-begin");
	value = strtok(attr, "npt=s");
	tempChar = (char *) malloc (strlen(value)+1);
	strcpy(tempChar, value);
	
	if (DEBUG_NCCSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clip-begin: %s", value);
		report (error, REP_DEBUG);
	}
	
	// storing clip-begin
	tmpNode->audioStartPos = tempChar;	
	if (attr != NULL) xmlFree (attr); attr = NULL;
	
	// fetch clip-end
	attr = xmlTextReaderGetAttribute (reader, "clip-end");
	value = strtok (attr, "npt=s");
	tempChar = (char *) malloc (strlen(value)+1);
	strcpy(tempChar, value);
	
	if (DEBUG_NCCSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clip-end: %s", value);
		report (error, REP_DEBUG);
	}
	
	// storing clip-end
	tmpNode->audioStopPos = tempChar;	
	if (attr != NULL) xmlFree (attr); attr = NULL;
	return ret;
}


/************************ END of *.smil parser code ************************/

/********************** Code for parsing a xhtml files **********************/


/**
 * Parses a xhtml-fil.
 * 
 * @param filename the filename to the xhtml-file.
 * @param fragment the fragment to fetch.
 */
static int parseXHTML (char* filename, char* fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
 	xmlTextReaderPtr reader = NULL;
	
	char tmp[STRLEN];
	strcpy(tmp, path);
	strcat(tmp, "/");
	strcat(tmp, filename);
	
	// open the xhtml file
 	reader = xmlReaderForFile (tmp, NULL, 0); 
	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open XHTML file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCCXHTML)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing XHTML file: %s", filename);
	 		report (error, REP_DEBUG);
		}
		
		ret = xmlTextReaderRead(reader);
		if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse XHTML file: %s", filename);
 			report (error, REP_WARNING);
 			xmlFreeTextReader (reader);
 			return -1;
	 	}
		while(ret == 1)
		{
			name = xmlTextReaderConstName (reader);		
			if (name == NULL) return 0;
			//parse tag if <head>	
			if (!strcmp (name, "head"))
			{
				ret = parseXHTMLHead (reader);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse XHTML file: %s", filename);
		 			report (error, REP_WARNING);
		 			xmlFreeTextReader (reader);
		 			return -1;
	 			}
			}
			//parse tag if <body>
			else if (!strcmp (name, "body"))
			{
				ret = parseXHTMLBody (reader, fragment, tmpNode);
				xmlFreeTextReader (reader);
				if (ret == 1) return 0; //TODO this function should return 1 hvis success and -1 if not
				else return ret;
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse XHTML file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return ret;
}

/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXHTMLHead (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCCXHTML) report ("Parsing XHTML <HEAD>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't come to the head closetag
	while (strcmp (name, "head"))
	{	
		//parse tag if <title>
		if (!strcmp (name, "title"))
		{
			ret = parseXHTMLTitle (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <meta>
		else if (!strcmp (name, "meta"))
		{
			ret = parseXHTMLMeta (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}
	return ret;
}

/**
 * Parse Title-tag. Extracts the title.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXHTMLTitle (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret = 1;
	
	if (DEBUG_NCCXHTML) report ("Parsing XHTML <TITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	// if the title tag has a value	
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces (value);
		//printf ("\t-%s-",tmp);
		//TODO: The title is never used..
	}
	if(value != NULL) xmlFree(value); value = NULL;
	if(tmp != NULL) free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}

/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXHTMLMeta (xmlTextReaderPtr reader){
	int ret = 1;
	if (DEBUG_NCCXHTML) report ("Parsing XHTML <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 * @param fragment the fragment to fetch.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXHTMLBody (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode)
{
	xmlChar *tmp, *value, *glue, part[500];
	tmp = value = glue = NULL; 
	char *temp = NULL;
	const xmlChar *name = NULL;	
	int ret = 1;
	
	if (DEBUG_NCCXHTML) report ("Parsing XHTML <BODY>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	while (ret == 1)
	{	
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
			
		// fetch the id-attribute
		tmp = xmlTextReaderGetAttribute (reader, "id");
		
			// if the current node has an id-attribute
			if (tmp != NULL) 
			{	
				// if the id-attribute equals the fragment
				if (!strcmp (tmp, fragment))
				{	
					name = xmlTextReaderConstName (reader);
					if (name == NULL) return 0;
					while (ret == 1) 
					{	
						ret = xmlTextReaderRead (reader);
						if (ret != 1) return ret;						
						if (!strcmp (name, xmlTextReaderConstName (reader)))
							ret = 0;
						else
						{						
							// get the text-value of the node
							if (ret == 1)
								value = xmlTextReaderValue (reader);
							else
								return ret;
							// if the value is not NULL
							if (value != NULL)
							{	
								//Temporary solution, get the text-value inside a possible
								//<b> tag, and glues it together with the previous text-value
								ret = xmlTextReaderRead (reader);
								if (ret == 1)
								{
									name = xmlTextReaderConstName (reader);
									if (name == NULL) return 0;
								}
								else
									return ret;
								//parse tag if <b>
								while (!strcmp (name, "b")) 
								{ 
									glue = parseXHTMLHtmltag(reader);
									if (glue != NULL)
									{
										part[0] = '\0';
										strcat (part, value);
										strcat (part, glue);
										if (value != NULL) xmlFree (value); value = NULL;
										if (glue != NULL) xmlFree (glue); glue = NULL;
										value = malloc (strlen (part)+1);	
										strcpy (value, part);
									}
									ret = xmlTextReaderRead (reader);
									if (ret != 1) return ret;
									ret = xmlTextReaderRead (reader);
									if (ret == 1)
									{
										name = xmlTextReaderConstName (reader);
										if (name == NULL) return 0;
									}
									else
										return ret;
									if (!strcmp (name, "#text")) {
										glue = xmlTextReaderValue (reader);
										temp = removewhitespaces(glue);
										part[0] = '\0';
										strcat (part, value);
										//strcat (part, " ");
										strcat (part, glue);
										if (value != NULL) xmlFree (value); value = NULL;
										if (temp != NULL) xmlFree (temp); temp = NULL;
										if (glue != NULL) xmlFree (glue); glue = NULL;
										value = malloc (strlen (part)+1);	
										strcpy (value, part);
									}
									ret = xmlTextReaderRead (reader);
									if (ret != 1) return ret;
									name = xmlTextReaderConstName (reader);
									if (name == NULL) return 0;
								}
								temp = removewhitespaces(value);
								if (temp != NULL)
								{
									tmpNode->textPassage = temp;
								}
								if (value != NULL) xmlFree (value); value = NULL;
								return ret;
							}
						}
					}	
				}
			}
		if (ret == 1)
		{
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
		}
	}
	if (tmp != NULL) xmlFree (tmp); tmp = NULL;
	return ret;
}


/**
 * Fetches the textvalue inside html-tags.
 * E.g: ....<b>text</b>...
 *
 * @param reader XmlTextReader.
 *
 * Return the textvalue.
 */
static char *parseXHTMLHtmltag (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	value = NULL;
	char *tempChar = NULL;
	
	xmlTextReaderRead (reader);
	//TODO: hva om det ikke er flere noder etter denne?
		
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tempChar = removewhitespaces(value);
		if (value != NULL) xmlFree (value); value = NULL;
	}
	return tempChar;	
}


/************************* END of xhtml parser code *************************/


#endif
