/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file ncxparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 *
 * NCX parser for the Daisy Z39.86-2005 standar. Collects data and stores it in linked list.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "ncxparser.h"
#include "control.h"
#include "report.h"
#include "common.h"

#ifdef LIBXML_READER_ENABLED

// static function declarations
static int parseNcxHead (xmlTextReaderPtr reader);
static int parseNcxDocTitle (xmlTextReaderPtr reader);
static int parseNcxDocAuthor (xmlTextReaderPtr reader);
static int parseNcxNavMap (xmlTextReaderPtr reader);
static int parseNcxPageList (xmlTextReaderPtr reader);
static int parseNcxNavList (xmlTextReaderPtr reader);
static int parseNcxSmilCustomTest (xmlTextReaderPtr reader);
static int parseNcxMeta (xmlTextReaderPtr reader);
static int parseNcxText (xmlTextReaderPtr reader);
static int parseNcxAudio (xmlTextReaderPtr reader);
static int parseNcxImg (xmlTextReaderPtr reader);
static int parseNcxNavInfo (xmlTextReaderPtr reader);
static int parseNcxNavLabel (xmlTextReaderPtr reader);
static int parseNcxNavPoint (xmlTextReaderPtr reader);
static int parseNcxContent (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseNcxPageTarget (xmlTextReaderPtr reader);
static int parseNcxNavTarget (xmlTextReaderPtr reader);
// global variables
extern char *path;


/************************ Code for parsing a NCX.* file ************************/


/**
 * Method for parsing of a NCX.* file
 * 
 * @param filname - the name of the ncc file
 * @returns 0 in case of success and -1 in case of error
 */
int parseNCX (char *filename)
{
 	const xmlChar *name;
 	xmlTextReaderPtr reader;
 	int ret;
 	
 	name = NULL;
 	// open the ncc file
 	reader = xmlReaderForFile (filename, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error),  "Failed to open NCX file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX file: %s", filename);
	 		report (error, REP_DEBUG);
		}
 		
	 	// read the first tag
	 	ret = xmlTextReaderRead(reader);
	 	
	 	// while there are more tags, parse the next
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse NCX file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
	 	}
	 	
	 	/* Attributes:
	 	 * version
	 	 * xmlns
	 	 * xml:lang
	 	 * dir
	 	 */
	 	
	 	while(ret == 1)
	 	{
	 		// get name of tag
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return ret;
			// parse tag if <head>
			if (!strcmp (name, "head")) 
				ret = parseNcxHead (reader);
			//parse tag if <docTitle>
			else if (!strcmp (name, "docTitle"))
				ret = parseNcxDocTitle (reader);
			//parse tag if <docAuthor> *
			else if (!strcmp (name, "docAuthor"))
				ret = parseNcxDocAuthor (reader);
			//parse tag if <navMap>
			else if (!strcmp (name, "navMap"))
				ret = parseNcxNavMap (reader);
			//parse tag if <pageList> ?
			else if (!strcmp (name, "pageList"))
				ret = parseNcxPageList (reader);
			//parse tag if <navList> *
			else if (!strcmp (name, "navList"))
				ret = parseNcxNavList (reader);
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse NCX file: %s", filename);
 			report (error, REP_WARNING
 			);
 			return -1;
		}
	}
	return 1;
}


/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxHead (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <HEAD>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't come to the head-close tag
	while (strcmp (name, "head"))
		{	
			//parse tag if <smilCustomTest>
			if (!strcmp (name, "smilCustomTest")) ret = parseNcxSmilCustomTest (reader);
			//parse tag if <meta>
			else if (!strcmp (name, "meta")) ret = parseNcxMeta (reader);
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);			
			if (name == NULL) return ret;
		}
	return ret;
}


/**
 * Parse SmilCustomTest-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxSmilCustomTest (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * id
	 * defaultState
	 * override
	 * bookStruct
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <SMILCUSTOMTEST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Meta-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxMeta (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * name
	 * content
	 * scheme
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse DocTitle-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocTitle (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	
	/* Attributes:
	 * id
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <DOCTITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	 
	// while we haven't reached the end-tag of doctitle
	while (strcmp (name, "docTitle"))
	{	
		//parse tag if <text>
		if (!strcmp (name, "text")) ret = parseNcxText (reader);
		//parse tag if <audio> ?
		else if (!strcmp (name, "audio")) ret = parseNcxAudio (reader);
		//parse tag if <img> ?
		else if (!strcmp (name, "img")) ret = parseNcxImg (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}	
	return ret;
}


/**
 * Parse Text-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxText (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <TEXT>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	// if the title tag has a value	
	if (xmlTextReaderHasValue (reader))
	{
		value = xmlTextReaderValue (reader);
		tmp = removewhitespaces (value);
		
	//	printf ("\n\t-%s-",tmp); //TODO: Store title of the book..
	}
	if (value != NULL) xmlFree (value); value = NULL;
	if (tmp != NULL) free(tmp); tmp = NULL; //TODO: do not free this when you put it in a data structure later on.
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Audio-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxAudio (xmlTextReaderPtr reader)
{
	xmlChar *tmp = NULL;
	char *tempChar;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * src
	 * clipBegin
	 * clipEnd
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <AUDIO>", REP_DEBUG); 
	// fetch the url anchor to the audio file
	tmp = xmlTextReaderGetAttribute (reader, "src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen(tmp)+1);
		strcpy(tempChar, tmp);
	//	tmpNode->audiofilename = tempChar; TODO add struct
	
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <AUDIO> Audiofile: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
		
		// fetch clipBegin
		tmp = xmlTextReaderGetAttribute (reader, "clipBegin");
		if (tmp != NULL)
		{
			tempChar = getTime(tmp);
	//	tmpNode->audiofilename = tempChar; TODO add struct
	
			if (DEBUG_NCXPARSER)
			{
				char error[STRLEN];
				snprintf (error, sizeof (error), "Parsing NCX <AUDIO> clipBegin: %s", tempChar);
				report (error, REP_DEBUG);
			}
			
			if (tmp != NULL) xmlFree (tmp); tmp = NULL;
						
			// fetch clip-end
			tmp = xmlTextReaderGetAttribute (reader, "clipEnd");
			if (tmp != NULL)
			{
				tempChar = getTime(tmp);
	//	tmpNode->audiofilename = tempChar; TODO add struct
	
				if (DEBUG_NCXPARSER)
				{
					char error[STRLEN];
					snprintf (error, sizeof (error), "Parsing NCXL <AUDIO> clipEnd: %s", tempChar);
					report (error, REP_DEBUG);
				}
			}
		}
		else
		{
			ret = 0;
		}
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}


/**
 * Parse Img-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxImg (xmlTextReaderPtr reader)
{
	xmlChar *tmp = NULL;
	char *tempChar;
	int ret = 1;

	/* Attributes:
	 * id
	 * class
	 * src
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <IMG>", REP_DEBUG);
	// fetch the url anchor to the audio file
	tmp = xmlTextReaderGetAttribute (reader, "src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen(tmp)+1);
		strcpy(tempChar, tmp);
	//	tmpNode->audiofilename = tempChar; TODO add struct
	
		if (DEBUG_NCXPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <IMG> Imagefile: %s", tmp);
			report (error, REP_DEBUG);
		}		
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}

/**
 * Parse DocAuthor-tag. 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxDocAuthor (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <DOCAUTHOR>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while the current node is text or audio or img
	while (strcmp (name, "docAuthor"))
	{	
		//parse tag if <text>
		if (!strcmp (name, "text")) ret = parseNcxText (reader);
		//parse tag if <audio> ?
		else if (!strcmp (name, "audio")) ret = parseNcxAudio (reader);
		//parse tag if <img> ?
		else if (!strcmp (name, "img")) ret = parseNcxImg (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}	
	return ret;
}


/**
 * Parse NavMap-tag 
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavMap (xmlTextReaderPtr reader)
{
	const xmlChar *name;
	int ret = 1;
	
	name = NULL;
		/* Attributes:
	 * 
	 */
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVMAP>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't reached the end-tag of navMap
	while (strcmp (name, "navMap"))
	{	
		//parse tag if <navInfo> *
		if (!strcmp (name, "navInfo")) ret = parseNcxNavInfo (reader);
		//parse tag if <navLabel> *
		else if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <navPoint> +
		else if (!strcmp (name, "navPoint")) ret = parseNcxNavPoint (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse PageList-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageList (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <PAGELIST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't come to the end-tag of pageList
	while (strcmp (name, "pageList"))
	{	
		//parse tag if <navInfo> *
		if (!strcmp (name, "navInfo")) ret = parseNcxNavInfo (reader);
		//parse tag if <navLabel> *
		else if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <pageTarget> +
		else if (!strcmp (name, "pageTarget")) ret = parseNcxPageTarget (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse NavList-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavList (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * class
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVLIST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't come to the end-tag of navList
	while (strcmp (name, "navList"))
	{	
		//parse tag if <navInfo> *
		if (!strcmp (name, "navInfo")) ret = parseNcxNavInfo (reader);
		//parse tag if <navLabel> +
		else if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <navTarget> +
		else if (!strcmp (name, "navTarget")) ret = parseNcxNavTarget (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}		
	return ret;
}


/**
 * Parse NavInfo-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavInfo (xmlTextReaderPtr reader)
{	
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * xml:lang
	 * dir
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVINFO>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't reached the end-tag of navInfo
	while (strcmp (name, "navInfo"))
	{	
		//parse tag if <text>
		if (!strcmp (name, "text")) ret = parseNcxText (reader);
		//parse tag if <audio> (if <text> then <audio> ?)
		else if (!strcmp (name, "audio")) ret = parseNcxAudio (reader);
		//parse tag if <img> ?
		else if (!strcmp (name, "img")) ret = parseNcxImg (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if(name == NULL) return 0;
	}	
	return ret;
}


/**
 * Parse NavLabel-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavLabel (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;

	/* Attributes:
	 * xml:lang
	 * dir
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVLABEL>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached the end-tag of navLabel
	while (strcmp (name, "navLabel"))
	{	
		//parse tag if <text>
		if (!strcmp (name, "text")) ret = parseNcxText (reader);
		//parse tag if <audio> (if <text> then <audio> ?)
		else if (!strcmp (name, "audio")) ret = parseNcxAudio (reader);
		//parse tag if <img> ?
		else if (!strcmp (name, "img")) ret = parseNcxImg (reader);
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}


/**
 * Parse NavPoint-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavPoint (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	/* Attributes:
	 * id
	 * class
	 * playOrder
	 */
	 
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVPOINT>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	struct SmilNode *tmpSmilNode = addNewSmilNode();
	
	// while the current node is navlabel, content, navpoint or #text
	// TODO if a tag is wrong the whole parsing is fubar..
	//		can't check if we have reached the end of navpoint since its recursive.
	while (!strcmp (name, "navLabel") ||  !strcmp (name, "content") ||  !strcmp (name, "navPoint") || !strcmp (name, "#text"))
	{	
		//parse tag if <navLabel> +
		if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <content>
		else if (!strcmp (name, "content")) ret = parseNcxContent (reader, tmpSmilNode);
		//parse tag if <navPoint> *
		else if (!strcmp (name, "navPoint")) ret = parseNcxNavPoint (reader);

		ret = xmlTextReaderRead (reader);
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		if(ret == 0) return ret;
	}	
	return ret;
}


/**
 * Parse Content-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxContent (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	xmlChar *smilsrc, *tmp;
	smilsrc = tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * src
	 */

	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVCONTENT>", REP_DEBUG);
	// gets the href attribute.
	smilsrc = xmlTextReaderGetAttribute (reader, "src");
	if (smilsrc != NULL)
	{
		// splits the text on #.
		tmp = strtok (smilsrc, "#");
		tempChar = (char *) malloc (strlen (tmp)+1);
		strcpy(tempChar, tmp);
		
		if (DEBUG_NCXPARSER)
		{ 
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <CONTENT> File anchor: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		// put the anchor in the node
		if (tmpSmilNode != NULL)
		tmpSmilNode->anchor = tempChar; tempChar = NULL;
		
		// get the fragment identifier
		tmp = strtok (NULL, "#");
		tempChar = (char *) malloc (strlen (tmp)+1);
		strcpy(tempChar, tmp);
		
		if (DEBUG_NCXPARSER)
		{ 
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing NCX <CONTENT> Fragment Identifier: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		// put the fragment identifier in the node
		if (tmpSmilNode != NULL)
		tmpSmilNode->fragmentIdentifier = tempChar; tempChar = NULL;
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		if (smilsrc != NULL) xmlFree(smilsrc); smilsrc = NULL;
	}
	else
	{
		ret = 0;
	}
	return ret;
}

/**
 * Parse PageTarget-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxPageTarget (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * value
	 * type
	 * class
	 * playOrder
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <PAGETARGET>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;
	
	// while we haven't come to the end tag of pageTarget
	while (!strcmp (name, "pageTarget"))
	{	
		//parse tag if <navLabel> +
		if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <content>
		else if (!strcmp (name, "content")) ret = parseNcxContent (reader, NULL);
		ret = xmlTextReaderRead (reader);
		name = xmlTextReaderConstName (reader);
		if(ret == 0) return ret;
	}	
	return ret;
}


/**
 * Parse NavTarget-tag
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseNcxNavTarget (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1; 
	
	/* Attributes:
	 * id
	 * value
	 * class
	 * playOrder
	 */
	
	if (DEBUG_NCXPARSER) report ("Parsing NCX <NAVTARGET>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return ret;

	// while we haven't come to the end tag of navTarget
	while (strcmp (name, "navTarget"))
	{
		//parse tag if <navLabel> +
		if (!strcmp (name, "navLabel")) ret = parseNcxNavLabel (reader);
		//parse tag if <content>
		else if (!strcmp (name, "content")) ret = parseNcxContent (reader, NULL);
		ret = xmlTextReaderRead (reader);
		name = xmlTextReaderConstName (reader);
		if(ret == 0) return ret;
	}		
	return ret;
}


/************************* END of NCX.* parser code *************************/

#endif
