/***************************************************************************
 *   Copyright (C) 2006 by André Lindhjem <belgarat@sdf.lonestar.org>,     *
 *                         Kjetil Holien <kjetil.holien@gmail.com>,        *
 *                         Terje Risa <terje.risa@gmail.com> &             *
 *                         Øyvind Nerbråten <oyvind@nerbraten.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************
 *
 * \file smilparser.c
 * \author André Lindhjem, Kjetil Holien, Terje Risa & Øyvind Nerbråten
 * \date 23.02.2006
 * 
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml2/libxml/xmlreader.h>

#include "ncxsmilparser.h"
#include "control.h"
#include "report.h"
#include "common.h"

#ifdef LIBXML_READER_ENABLED

// static function declarations 
static int parseSmilHead (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilLayout (xmlTextReaderPtr reader);
static int parseSmilMeta (xmlTextReaderPtr reader);
static int parseSmilRegion (xmlTextReaderPtr reader);
static int parseSmilCustomAttributes (xmlTextReaderPtr reader);
static int parseSmilCustomTest (xmlTextReaderPtr reader);
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode);
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilImg (xmlTextReaderPtr reader, struct Node *tmpNode);
static int parseSmilA (xmlTextReaderPtr reader);
static int parseXML (char* filename, char* fragment, struct Node *tmpNode);
static int parseXMLHead (xmlTextReaderPtr reader);
static int parseXMLTitle (xmlTextReaderPtr reader);
static int parseXMLMeta (xmlTextReaderPtr reader);
static int parseXMLFrontmatter (xmlTextReaderPtr reader);
static int parseXMLBodymatter (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode);
static int parseXMLRearmatter (xmlTextReaderPtr reader);
static int parseXMLDoctitle (xmlTextReaderPtr reader);
static int parseXMLDocauthor (xmlTextReaderPtr reader);
static int parseXMLLevel (xmlTextReaderPtr reader);
static int parseXMLLevel1 (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode);
static int parseXMLBlock (xmlTextReaderPtr reader);
static int parseXMLBook (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode);
//static char* parseXMLHtmltag (xmlTextReaderPtr reader);


// global variables
static int alreadyfound = 0; //Used in the smil-parsing
extern char *path;


/************************ Code for parsing a *.smil files ************************/


/**
 * Method for parsing of a *.smil file
 * 
 * @param tmpSmilNode - the smilNode for the smilfile to parse.
 * @returns 1 in case of success and -1 in case of error
 */
int parseNCXSmil (struct SmilNode *tmpSmilNode)
{
 	const xmlChar *name = NULL;
 	xmlTextReaderPtr reader = NULL;
 	int ret = 1;
 	
  	char tmp[STRLEN];
	strcpy(tmp, path);
	strcat(tmp, "/");
	strcat(tmp, tmpSmilNode->anchor);

 	// open the smil file
 	reader = xmlReaderForFile (tmp, NULL, 0);
 	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open SMIL file: %s", tmpSmilNode->anchor);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL file: %s", tmpSmilNode->anchor);
	 		report (error, REP_DEBUG);
		}
 			
	 	// read the first tag
	 	ret = xmlTextReaderRead(reader);
	 	if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			return -1;
	 	}
	 	// while there are more tags, parse the next
	 	while(ret == 1)
	 	{
			/* Attributes:
	 		 * id
	 		 * class
	 		 * title
			 * xmlns
			 * xml:lang
			 */
	 
	 		// get name of tag
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
			//parse tag if <head>
			if (!strcmp (name, "head"))
			{ 
				ret = parseSmilHead (reader, tmpSmilNode);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}
			}
			//parse tag if <body>
			else if (!strcmp (name, "body"))
			{ 
				ret = parseSmilBody (reader, tmpSmilNode);
				if (ret != 0)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse SMIL file: %s", tmpSmilNode->anchor);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return 1;
}


/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilHead(xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <HEAD>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached the head-close tag
	while (strcmp (name, "head"))
	{	
		//parse tag if <layout> ?
		if (!strcmp (name, "layout"))
		{
			ret = parseSmilLayout (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <meta> *
		else if (!strcmp (name, "meta"))
		{
			ret = parseSmilMeta (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <customAttributes> ?
		else if (!strcmp (name, "customAttributes"))
		{
			ret = parseSmilCustomAttributes (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Layout-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilLayout (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <LAYOUT>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached the layout-close tag
	while (strcmp (name, "layout"))
	{	
		//parse tag if <region> +
		if (!strcmp (name, "region"))
		{
			ret = parseSmilRegion (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilMeta (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * content
	 * name
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Region-tag. Currently not implemented
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilRegion (xmlTextReaderPtr reader)
{
	int ret = 1;
	
	/* Attributes:
	 * id
	 * bottom
	 * left
	 * right
	 * top
	 * height
	 * width
	 * fit
	 * backgroundColor
	 * showBackground
	 * z-index
	 * xml:lang
	 */
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <REGION>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse CustomAttributes-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilCustomAttributes (xmlTextReaderPtr reader)
{	
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <CUSTOMATTRIBUTES>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached the customattributes-close tag
	while (strcmp (name, "customAttributes"))
	{	
		//parse tag if <customTest> +
		if (!strcmp (name, "customTest"))
		{
			ret = parseSmilCustomTest (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}


/**
 * Parse CustomTest-tag. Currently not implemented
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilCustomTest (xmlTextReaderPtr reader)
{	
	int ret;
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 * defaultState
	 * override
	 */

	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <CUSTOMTEST>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}


/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilBody (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 */

	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <BODY>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached that body close tag
	while (strcmp (name, "body"))
	{	
		//parse tag if <par> +
		if (!strcmp (name, "par"))
		{
			ret = parseSmilPar (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <seq> +
		else if (!strcmp (name, "seq"))
		{
			ret = parseSmilSeq (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <text> +
		else if (!strcmp (name, "text"))
		{
			ret = parseSmilText (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <audio> +
		else if (!strcmp (name, "audio"))
		{
			ret = parseSmilAudio (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <img> +
		else if (!strcmp (name, "img"))
		{
			ret = parseSmilImg (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <a> +
		else if (!strcmp (name, "a"))
		{
			ret = parseSmilA (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Main-Seq-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilSeq (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * class
	 * customTest
	 * dur
	 * xml:lang
	 * end
	 * fill
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <SEQ>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while the current node is par, seq, text, audio, img, a or #text
	while (!strcmp (name, "par") ||  !strcmp (name, "seq") || !strcmp (name, "text")
		||  !strcmp (name, "audio") ||  !strcmp (name, "img") ||  !strcmp (name, "a") || !strcmp (name, "#text"))
	{	
		//parse tag if <par> +
		if (!strcmp (name, "par"))
		{
			ret = parseSmilPar (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <seq> +
		else if (!strcmp (name, "seq"))
		{
			ret = parseSmilSeq (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <text> +
		else if (!strcmp (name, "text"))
		{
			ret = parseSmilText (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <audio> +
		else if (!strcmp (name, "audio"))
		{
			ret = parseSmilAudio (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <img> +
		else if (!strcmp (name, "img"))
		{
			ret = parseSmilImg (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <a> +
		else if (!strcmp (name, "a"))
		{
			ret = parseSmilA (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}

/**
 * Parse Par-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilPar (xmlTextReaderPtr reader, struct SmilNode *tmpSmilNode)
{
	const xmlChar *name = NULL;
	xmlChar *tmp = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * customTest
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <PAR>", REP_DEBUG);
	// go to the fragment identifier point and parses from there.
	tmp = xmlTextReaderGetAttribute (reader, "id");
		
	// if the par tag has an id attribute, read it and check wheter it matches the fragmentid.
	// if there is none id attribute, set alreadyfound to 1 and parse the whole smilfile
	if (tmp != NULL) 
	{
	//	if (tmpSmilNode->fragmentIdentifier != NULL)
		//{
			//if (!strcmp (tmp, tmpSmilNode->fragmentIdentifier))
			alreadyfound = 1;
		//}
	}
	else
		alreadyfound = 1;
	// should we continue parsing if the id tag is missing?
	// the way it is now, if the id tag is missing we parse from here and out
	if (tmp != NULL) xmlFree (tmp); tmp = NULL;
	if (alreadyfound == 1)
	{
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		struct Node *tmpNode = addNewNode();
		
		// while we haven't reached the par-end tag
		while (strcmp (name, "par"))
		{	
			//parse tag if <seq> +
			if (!strcmp (name, "seq"))
			{
				ret = parseSmilSeq (reader, tmpSmilNode);
				if (ret != 1) return ret;
			}
			//parse tag if <text> +
			else if (!strcmp (name, "text"))
			{
				ret = parseSmilText (reader, tmpNode);
				if (ret != 1) return ret;
			}
			//parse tag if <audio> +
			else if (!strcmp (name, "audio"))
			{
				ret = parseSmilAudio (reader, tmpNode);
				if (ret != 1) return ret;
			}
			//parse tag if <img> +
			else if (!strcmp (name, "img"))
			{
				ret = parseSmilImg (reader, tmpNode);
				if (ret != 1) return ret;
			}
			//parse tag if <a> +
			else if (!strcmp (name, "a"))
			{
				ret = parseSmilA (reader);
				if (ret != 1) return ret;
			}
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
		}
	}
	else 
	{
		ret = xmlTextReaderNext (reader);	
	}
	return ret;
}


/**
 * Parse Text-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilText (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *textsrc, *fragment, *attr;
	textsrc = fragment = attr = NULL;
	const xmlChar *name = NULL;
	char *tmp = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * src
	 * type
	 * region
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <TEXT>", REP_DEBUG);
	name = xmlTextReaderConstName (reader);
	if (!strcmp(name, "#text")) {
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
	}
	attr = xmlTextReaderGetAttribute(reader, "src");	
	
	// splits the text on #.
	// get the anchor url
	textsrc = strtok (attr, "#");
	
	if (DEBUG_NCXSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> XMLFile: %s", textsrc);
		report (error, REP_DEBUG);
	}
	
	// storing the filename of the text file
	tmp = (char *) malloc (strlen (textsrc)+1);
	strcpy (tmp, textsrc);
	tmpNode->textfilename = tmp;
	
	// get the fragment identifier
	fragment = strtok (NULL, "#");
	
	if (DEBUG_NCXSMILPARSER)
	{
		char error[STRLEN];
		snprintf (error, sizeof (error), "Parsing SMIL <TEXT> Fragment Identifier: %s", fragment);
		report (error, REP_DEBUG);
	}
	
	// storing the fragment identifier for the passage in the text file
	tmp = (char *) malloc (strlen (fragment)+1);
	strcpy (tmp, fragment);
	tmpNode->fragmentidentifier = tmp;
	
	// parse text file
	if (parseXML(textsrc, fragment, tmpNode) != 0)
	{
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse XHTML file: %s", tmpNode->textfilename);
	 		report (error, REP_WARNING);
		}
 		return -1;
	}
	
	if (attr != NULL) xmlFree (attr);
	return ret;
}

/**
 * Parse Audio-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilAudio (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;

	/* Attributes:
	 * id
	 * src
	 * type
	 * clipBegin
	 * clipEnd
	 * region
	 * xml:lang
	 */
	// fetch the url anchor to the audio file
	tmp = xmlTextReaderGetAttribute (reader, "src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen(tmp)+1);
		strcpy(tempChar, tmp);
		tmpNode->audiofilename = tempChar;
		
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> Audiofile: %s", tmp);
			report (error, REP_DEBUG);
		}
		
		if (tmp != NULL) xmlFree (tmp); tmp = NULL;
		
		// fetch clipBegin
		tmp = xmlTextReaderGetAttribute (reader, "clipBegin");
		if (tmp != NULL)
		{
			tempChar = getTime(tmp);
			tmpNode->audiofilename = tempChar;
			
			if (DEBUG_NCXSMILPARSER)
			{
				char error[STRLEN];
				snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clipBegin: %s", tempChar);
				report (error, REP_DEBUG);
			}
			
			if (tmp != NULL) xmlFree (tmp); tmp = NULL;
			
			// fetch clipEnd
			tmp = xmlTextReaderGetAttribute (reader, "clipEnd");
			if (tmp != NULL)
			{
				tempChar = getTime(tmp);
				tmpNode->audiofilename = tempChar;
				
				if (DEBUG_NCXSMILPARSER)
				{
					char error[STRLEN];
					snprintf (error, sizeof (error), "Parsing SMIL <AUDIO> clipEnd: %s", tempChar);
					report (error, REP_DEBUG);
				}
			}
		}
		else
		{
			ret = 0;
		}
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp); tmp = NULL;	
	return ret;
}


/**
 * Parse Img-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilImg (xmlTextReaderPtr reader, struct Node *tmpNode)
{
	xmlChar *tmp = NULL;
	char *tempChar = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * src
	 * type
	 * region
	 * xml:lang
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <IMG>", REP_DEBUG);
	// fetch the url anchor to the audio file
	tmp = xmlTextReaderGetAttribute (reader, "src");
	if (tmp != NULL)
	{
		tempChar = (char *) malloc (strlen(tmp)+1);
		strcpy(tempChar, tmp);
	//	tmpNode->imagefilename = tempChar; TODO add struct
	
		if (DEBUG_NCXSMILPARSER)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing SMIL <IMG> Imagefile: %s", tmp);
			report (error, REP_DEBUG);
		}	
	}
	else
	{
		ret = 0;	
	}
	if (tmp != NULL) xmlFree (tmp);	tmp = NULL;
	return ret;
}


/**
 * Parse A-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseSmilA (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	
	/* Attributes:
	 * id
	 * class
	 * title
	 * xml:lang
	 * href
	 * external
	 */
	 
	if (DEBUG_NCXSMILPARSER) report ("Parsing SMIL <A>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;
	
	// while we haven't reached the a-close tag
	while (strcmp (name, "a"))
	{	
		//parse tag if <par> *
		if (!strcmp (name, "par"))
		{
			ret = parseSmilPar (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <seq> *
		else if (!strcmp (name, "seq"))
		{
			ret = parseSmilSeq (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <text> *
		else if (!strcmp (name, "text"))
		{
			ret = parseSmilText (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <audio> *
		else if (!strcmp (name, "audio"))
		{
			ret = parseSmilAudio (reader, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <img> *
		else if (!strcmp (name, "img"))
		{
			ret = parseSmilImg (reader, NULL);
			if (ret != 1) return ret;
		}		
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
	}	
	return ret;
}
/************************ END of *.smil parser code ************************/

/********************** Code for parsing a XML files **********************/


/**
 * Parses a xml-fil.
 * 
 * @param filename the filename to the xhtml-file.
 * @param fragment the fragment to fetch.
 */
static int parseXML (char* filename, char* fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;
	int ret = 1;
 	xmlTextReaderPtr reader = NULL;	
	
	char tmp[STRLEN];
	strcpy(tmp, path);
	strcat(tmp, filename);

	// open the xml file
 	reader = xmlReaderForFile (tmp, NULL, 0); 
	if (reader == NULL)
 	{
 		char error[STRLEN];
 		snprintf (error, sizeof (error), "Failed to open XML file: %s", filename);
 		report (error, REP_WARNING);
 		return -1;
 	}
 	else 
	{
		if (DEBUG_NCXXML)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Parsing XML file: %s", filename);
	 		report (error, REP_DEBUG);
		}
		
		ret = xmlTextReaderRead(reader);
		if (ret != 1)
	 	{
	 		char error[STRLEN];
	 		snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
	 	}
		while(ret == 1)
		{	
			/* Attributes:
	 		 */
	 		 
			name = xmlTextReaderConstName (reader);	
			//parse tag if <head>
			if (!strcmp (name, "head"))
			{
				ret = parseXMLHead (reader);
				if (ret != 1)
				{
					char error[STRLEN];
			 		snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
		 			report (error, REP_WARNING);
		 			return -1;
			 	}
			}
			//parse tag if <book>
			else if (!strcmp (name, "book"))
			{
				ret = parseXMLBook (reader, fragment, tmpNode);
				xmlFreeTextReader (reader);
				if (ret == 1) return 0;
				else return ret;
			}
			ret = xmlTextReaderRead (reader);
		}
		xmlFreeTextReader (reader);
		if (ret != 0)
		{
			char error[STRLEN];
			snprintf (error, sizeof (error), "Failed to parse XML file: %s", filename);
 			report (error, REP_WARNING);
 			return -1;
		}
	}
	return ret;
}

/**
 * Parse Head-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLHead (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XHTML <HEAD>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while we haven't reached the head-close tag
	while (strcmp (name, "head"))
	{	
		//parse tag if <title>
		if (!strcmp (name, "title"))
		{
			ret = parseXMLTitle (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <meta>
		else if (!strcmp (name, "meta"))
		{
			ret = parseXMLMeta (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	return ret;
}

/**
 * Parse Title-tag. Extracts the title.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLTitle (xmlTextReaderPtr reader)
{
	xmlChar *value = NULL;
	char *tmp = NULL;
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <TITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	// if the title tag has a value	
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces (value);
//		printf ("\t-%s-",tmp);
		//TODO: The title is never used..
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}

/**
 * Parse Meta-tag. Currently not implemented.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLMeta (xmlTextReaderPtr reader){
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <META>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	return ret;	
}

/**
 * Parse Body-tag.
 *
 * @param reader XmlTextReader.
 * @param fragment the fragment to fetch.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLBook (xmlTextReaderPtr reader, char *fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <BOOK>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while we haven't reached the book-close tag
	while (strcmp (name, "book"))
	{	
		//parse tag if <frontmatter>
		if (!strcmp (name, "frontmatter"))
		{
			ret = parseXMLFrontmatter (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <bodymatter>
		else if (!strcmp (name, "bodymatter"))
		{
			ret = parseXMLBodymatter (reader, fragment, tmpNode);
			if (ret == 1) return ret;
			//if (ret != 1) return ret;
		}
		//parse tag if <rearmatter>
		else if (!strcmp (name, "rearmatter"))
		{
			ret = parseXMLRearmatter (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	return ret;
}


/**
 * Parse Frontmatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLFrontmatter (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <FRONTMATTER>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while we haven't reached the frontmatter-close tag
	while (strcmp (name, "frontmatter"))
	{	
		//parse tag if <doctitle>
		if (!strcmp (name, "doctitle"))
		{
			ret = parseXMLDoctitle (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <docauthor>
		else if (!strcmp (name, "docauthor"))
		{
			ret = parseXMLDocauthor (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <level>
		else if (!strcmp (name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <level1>
		else if (!strcmp (name, "level1"))
		{
			ret = parseXMLLevel1 (reader, NULL, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <block>
		else if (!strcmp (name, "block"))
		{
			ret = parseXMLBlock (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	return ret;
}


/**
 * Parse Bodymatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLBodymatter (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <BODYMATTER>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while we haven't reached the bodymatter-close tag
	while (strcmp (name, "bodymatter"))
	{	
		//parse tag if <level>
		if (!strcmp (name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <level1>
		else if (!strcmp (name, "level1"))
		{
			ret = parseXMLLevel1 (reader, fragment, tmpNode);
			if (ret == 1) return ret;
		//	if (ret != 1) return ret;
		}
		//parse tag if <block>
		else if (!strcmp (name, "block"))
		{
			ret = parseXMLBlock (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	return ret;
}


/**
 * Parse Rearmatter-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLRearmatter (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <REARMATTER>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	
	// while we haven't reached the rearmatter-close tag
	while (strcmp (name, "rearmatter"))
	{	
		//parse tag if <level>
		if (!strcmp (name, "level"))
		{
			ret = parseXMLLevel (reader);
			if (ret != 1) return ret;
		}
		//parse tag if <level1>
		else if (!strcmp (name, "level1"))
		{
			ret = parseXMLLevel1 (reader, NULL, NULL);
			if (ret != 1) return ret;
		}
		//parse tag if <block>
		else if (!strcmp (name, "block"))
		{
			ret = parseXMLBlock (reader);
			if (ret != 1) return ret;
		}
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		name = xmlTextReaderConstName (reader);
	}
	return ret;
}


/**
 * Parse Doctitle-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLDoctitle (xmlTextReaderPtr reader)
{	
	xmlChar *value, *tmp;
	value = tmp = NULL;
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <DOCTITLE>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	// if the title tag has a value	
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces (value);
		//printf ("\t-%s-",tmp);
		//TODO: The title is never used..
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Docauthor-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLDocauthor (xmlTextReaderPtr reader)
{
	xmlChar *value, *tmp;
	value = tmp = NULL;
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <DOCAUTHOR>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	
	// if the title tag has a value	
	if (xmlTextReaderHasValue (reader)){
		value = xmlTextReaderValue (reader);
		tmp =  removewhitespaces (value);
		//printf ("\t-%s-",tmp);
		//TODO: The title is never used..
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL) free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Level-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLLevel (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <LEVEL>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;	
	return ret;
}


/**
 * Parse Level1-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLLevel1 (xmlTextReaderPtr reader, char* fragment, struct Node *tmpNode)
{
	xmlChar *value, *tmp;
	value = tmp = NULL;
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <LEVEL1>", REP_DEBUG);
	
	// if the fragment = NULL - we are in frontmatter	
	if (fragment == NULL)
	{	
		ret = xmlTextReaderRead (reader);
		if (ret != 1) return ret;
		if(xmlTextReaderHasValue (reader))
		{
			value = xmlTextReaderValue (reader);
			tmp =  removewhitespaces (value);
			//printf ("\t-%s-",tmp);
			//TODO: The title is never used..
		}
	}
	// we are in bodymatter
	else
	{	
		// check wheter it is the level1 tag who got the id we request
		tmp = xmlTextReaderGetAttribute (reader, "id");
		ret = xmlTextReaderRead (reader);
		if (tmp != NULL)
		{	
			if (!strcmp (tmp, fragment))
			{
				while(!xmlTextReaderHasValue (reader))
				{
					ret = xmlTextReaderRead (reader);
					
					/* check wheter we have come to a new tag with id,
					 * if that is the case, the level1 tag didn't have
					 * any valid values 
					 */
					tmp = xmlTextReaderGetAttribute (reader, "id");
					if (tmp != NULL) return 0;
				}
				value = xmlTextReaderValue (reader);
				tmp =  removewhitespaces (value);
				if(tmp != NULL) tmpNode->textPassage = tmp;
				if(value != NULL) xmlFree(value); value = NULL;	
				printf ("\n%s\n",tmp);
				return 1;	// return succesful reading	
			}
		}
		name = xmlTextReaderConstName (reader);
		if (name == NULL) return 0;
		
		// while we haven't reached the level1-endtag
		while (strcmp (name, "level1"))
		{	
			tmp = xmlTextReaderGetAttribute (reader, "id");
			if (tmp != NULL)
			{	
				if (!strcmp (tmp, fragment))
				{	ret = xmlTextReaderRead (reader);
					if(xmlTextReaderHasValue (reader))
					{
						value = xmlTextReaderValue (reader);
						tmp =  removewhitespaces (value);
						if(tmp != NULL) tmpNode->textPassage = tmp;
						if(value != NULL) xmlFree(value); value = NULL;										
						return 1;	
					}
					else
					{
						if(value != NULL) xmlFree(value); value = NULL;										
						if(tmp != NULL) xmlFree(tmp); tmp = NULL;																
						return 0;
					}
				}				
			}
			ret = xmlTextReaderRead (reader);
			if (ret != 1) return ret;
			name = xmlTextReaderConstName (reader);
			if (name == NULL) return 0;
		}		
	}
	if(value != NULL) 
		xmlFree(value); value = NULL;
	if(tmp != NULL)
		free(tmp); tmp = NULL;
	ret = xmlTextReaderRead (reader);
	return ret;
}


/**
 * Parse Block-tag.
 *
 * @param reader XmlTextReader.
 *
 * Return 1 if the node is read successfully, 0 if there is no more nodes to read, -1 in case of error.
 */
static int parseXMLBlock (xmlTextReaderPtr reader)
{
	const xmlChar *name = NULL;	
	int ret = 1;
	/* Attributes:
	 */
	if (DEBUG_NCXXML) report ("Parsing XML <BLOCK>", REP_DEBUG);
	ret = xmlTextReaderRead (reader);
	if (ret != 1) return ret;
	name = xmlTextReaderConstName (reader);
	if (name == NULL) return 0;	
	return ret;
}


///**
// * Fetches the textvalue inside html-tags.
// * E.g: ....<b>text</b>...
// *
// * @param reader XmlTextReader.
// *
// * Return the textvalue.
// */
//static char *parseXMLHtmltag (xmlTextReaderPtr reader)
//{
//	xmlChar *value = NULL;
//	value = NULL;
//	char *tempChar = NULL;
//	
//	xmlTextReaderRead (reader);
//	//TODO: hva om det ikke er flere noder etter denne?
//		
//	if (xmlTextReaderHasValue (reader))
//	{
//		value = xmlTextReaderValue (reader);
//		tempChar = removewhitespaces(value);
//		xmlFree (value);
//	}
//	return tempChar;	
//}


/************************* END of XML parser code *************************/


#endif
